import numpy as np
import numpy_financial as npf
from scipy import optimize
import warnings


def pv(irr: float, np_pmt, root=0):
    """
    Compute the present value of a stream of cash flows.
    Motivation: Unlike numpy_financial.npv, numpy_financial.pv doesn't support
        an arbitrary stream of cash flows.

    Parameters
    ----------
    irr : float
        Internal rate of return
    np_pmt : array_like
        A stream of cash flows
    root : float
        Used to invert the equation via optimize.root

    Returns
    -------
    float
        The present value of a stream of cash flows.
    """
    np_pmt = np.insert(np_pmt, 0, 0)
    pv = npf.npv(irr, np_pmt)
    return pv - root


def irr(presentvalue, np_pmt, use_scipy_optimize_root=True):
    """
    Return the Internal Rate of Return (IRR) of a stream of cash flows.

    Parameters
    ----------
    presentvalue : float
        The known present value of a stream of cash flows.
    np_pmt : array_like
        A stream of cash flows.
    use_scipy_optimize_root : bool
        If True, solvse for the IRR using scipy.optimize.root instead of
        numpy_financial.irr.

    Returns
    -------
    float
        Internal Rate of Return a stream of cash flows.

    Notes
    -----
    Note that the API is different than numpy_financial.irr.  Contains an option
    to solve for the IRR using scipy.optimize.root rather than
    numpy_financial.irr.  scipy.optimize.root with df-sane is typically much
    faster than numpy_financial.irr.
    """
    import warnings

    if use_scipy_optimize_root:
        with warnings.catch_warnings():
            # Original warning: numpy_financial\_financial.py:781:
            # RuntimeWarning: overflow encountered in power
            # return (values / (1+rate)**np.arange(0, len(values))).sum(axis=0)
            warnings.simplefilter('ignore')
            result = optimize.root(pv, 0, args=(np_pmt, presentvalue), method='df-sane')
        return result.x
    else:
        np_pmt = np.insert(np_pmt, 0, -presentvalue)
        irr = npf.irr(np_pmt)
        return irr


def mortgage_price(beyield, coupon, upb, np_pmt, dayofmonth=1,delay=47, root=0):
    """
    Compute the price of a mortgage loan or mortgage pool for an arbitary set
    of cash flows.

    Parameters
    ----------
    beyield : float
        The bond equivalent yield of a mortgage cash flow stream.
    coupon : float
        The annualized coupon of the pool.
    upb : float
        The unpaid principal balance of the mortgage pool.
    np_pmt : array_like
        A stream of monthly cash flows.
    dayofmonth : int
        The setlement date's day of month for clean price calculations.
    delay : int
        The cash flow delay assumption common in mortgage pool and mortgage
        backed security pricing.
    root : float
        Used to invert the equation via optimize.root.

    Returns
    -------
    float
        The price of the mortgage pool.

    Notes
    -----
    47 day delay is the market standard for loan trades according to Goldman
    Sachs' trading desk
    """
    monthly_coupon = coupon/1200
    irr = ((1+beyield/200)**(1/6)-1)
    dirty_price = pv(irr, np_pmt)/upb*100
    forward_price = dirty_price*(1+irr)**((dayofmonth-1+30-delay)/30)
    clean_price = forward_price - monthly_coupon*(dayofmonth-1)/360
    return clean_price - root


def mortgage_yield(price, coupon, upb, np_pmt, dayofmonth=1, delay=47):
    """
    Compute the price of a mortgage loan or mortgage pool for an arbitary set
    of cash flows.

    Parameters
    ----------
    price : float
        The price of a mortgage cash flow stream.
    coupon : float
        The annualized coupon of the pool.
    upb : float
        The unpaid principal balance of the mortgage pool.
    np_pmt : array_like
        A stream of monthly cash flows.
    dayofmonth : int
        The setlement date's day of month for clean price calculations.
    delay : int
        The cash flow delay assumption common in mortgage pool and mortgage
        backed security pricing.

    Returns
    -------
    float
        The price of the mortgage pool.

    Notes
    -----
    47 day delay is the market standard for loan trades according to Goldman
    Sachs' trading desk
    """
    result = optimize.root(mortgage_price, coupon,
                           args=(coupon, upb, np_pmt, dayofmonth, delay, price),
                           method='df-sane')
    return result.x



# from magicpandas.utils import search_iterable
from magicpandas.frame import MagicDataFrame
import os
import warnings
from contextlib import suppress
from magicpandas.easy_xlsxwriter import easy_xlsxwriter
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    import statsmodels.api as sm

def test_browse_in_excel():
    """Test that MagicDataFrame.browse creates an xlsx file in the temp directory"""
    import tempfile
    path_to_check = os.path.join(tempfile.gettempdir(), 'easy_xlsxwriter.xlsx')
    with suppress(FileNotFoundError):
        os.remove(path_to_check)
    df = sm.datasets.get_rdataset("Duncan", "carData").data
    df = MagicDataFrame(df)
    df.browse(client='excel', run=False)
    assert os.path.exists(path_to_check)

def test_browse_in_webbrowser():
    """Test that MagicDataFrame.browse creates an xlsx file in the temp directory"""
    import tempfile
    path_to_check = os.path.join(tempfile.gettempdir(), 'magicpandas_chromify.html')
    with suppress(FileNotFoundError):
        os.remove(path_to_check)
    df = sm.datasets.get_rdataset("Duncan", "carData").data
    df = MagicDataFrame(df)
    df.browse(client='webbrowser', run=False)
    assert os.path.exists(path_to_check)



from __future__ import annotations
import subprocess
import os
import re
import warnings
import numpy as np
import pandas as pd
import functools
from magicpandas.utils import search_iterable
from magicpandas.easy_xlsxwriter import easy_xlsxwriter
import altair as alt
from magicpandas.chromify import chromify
from magicpandas.finance.bc35 import bc35_price, bc35_yield
import inspect


def preserve_magic_attributes(func, magic_attributes=[
    'verbose_labels'
]):
    """Decorator to preserve DataFrame object attributes when running a
    MagicDataFrame method."""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # if isinstance(self, MagicDataFrame) or isinstance(self, MagicSeries):
        if issubclass(self.__class__, MagicDataFrame) or issubclass(self.__class__, MagicSeries):
            result = func(self, *args, **kwargs)
            # if isinstance(result, MagicDataFrame) or isinstance(result, MagicSeries):
            if issubclass(result.__class__, MagicDataFrame) or isinstance(result.__class__, MagicSeries):
                for magic_attribute in magic_attributes:
                    if hasattr(self, magic_attribute):
                        with warnings.catch_warnings():
                            warnings.simplefilter('ignore')
                            setattr(result, magic_attribute, getattr(self, magic_attribute))
            return result
        else:
            return func(self, *args, **kwargs)
    return wrapper


def for_all_methods(decorator, exclude=[None]):
    """A class decorator that decorates all methods with the same decorator,
    excluding dunder methods.

    References
    ----------
    https://stackoverflow.com/questions/6307761/how-to-decorate-all-functions-of-a-class-without-typing-it-over-and-over-for-eac
    """
    def decorate(cls):
        for attr in inspect.getmembers(cls, inspect.isfunction):
            # print(attr[0])
            # if attr[0] not in exclude:
            if attr[0] not in exclude and not re.match('^__.*__$', attr[0]):
                setattr(cls, attr[0], decorator(getattr(cls, attr[0])))
        return cls
    return decorate


class MagicSeries(pd.Series):
    @property
    def _constructor(self):
        return MagicSeries

    @property
    def _constructor_expanddim(self):
        return MagicDataFrame

    def pcut(self, step_size: float = None, bin_count: int = None,
             right: bool = True, format: str = '{:3.2f}') -> MagicSeries:
        """Pretty cut
        Converts MagicSeries of dtype float or int into a ordinal MagicSeries of
        dtype category that is formatted in a manner suitable for publishing in
        non-scientific contexts e.g., "1.001 - 2.000" vs (1.001, 2.000].

        Parameters
        ----------
        step_size : float
            The desired number of steps.  Note: must be None if bin_count is
            not None.
        bin_count : int
            The desired number of bins.  Note: must be None if step_size is
            not None.
        right : bool
            Indicates whether bins includes the rightmost edge or not.
        format : str
            Python style format string e.g., {:3.2f}

        Returns
        -------
        Categorical MagicSeries
            A MagicSeries representing the respective bin for each value of self.
        """
        if not self.dtype == 'float' and not self.dtype == 'int':
            raise TypeError('MagicSeries must of dtype float or int')

        if step_size is not None and bin_count is not None:
            raise ValueError('step_size is not None and bin_count is not None')
        elif step_size is None and bin_count is None:
            raise ValueError('step_size is None and bin_count is None')

        if bin_count is not None:
            step_size = (self.max() - self.min()) / (bin_count-1)
            step_size = round(step_size, 10)

        thisformat = format.replace('{:', '{0:') + ' - ' + format.replace('{:', '{1:')  # e.g., "{0:3.1f} - {1:3.1f}"
        delta = 1/10**int(re.findall(r'\.\d+', format)[0].replace('.', ''))  # e.g., 0.01

        minvalue = int(self.min()/step_size)*step_size
        maxvalue = int(self.max()/step_size+1)*step_size

        if right:
            bin_labels = [thisformat.format(i + delta, i + step_size) for i in
                          np.arange(minvalue, maxvalue, step_size)]
            bin_labels[0] = thisformat.format(minvalue, minvalue+step_size)
        else:
            bin_labels = [thisformat.format(i, i + step_size - delta) for i in
                          np.arange(minvalue, maxvalue, step_size)]

        return pd.cut(self, np.arange(minvalue, maxvalue+step_size, step_size),
                      right=right, labels=bin_labels, include_lowest=True)


# @for_all_methods(preserve_magic_attributes, ['__init__'])
@for_all_methods(preserve_magic_attributes)
class MagicDataFrame(pd.DataFrame):
    ############################################################################
    # methods to extend pandas.DataFrame
    ############################################################################
    @property
    def _constructor(self):
        return MagicDataFrame

    @property
    def _constructor_sliced(self):
        return MagicSeries

    ############################################################################
    # methods to set MagicDataFrame attributes e.g., verbose_labels
    ############################################################################
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not hasattr(self, 'verbose_labels'):
            with warnings.catch_warnings():
                # warning message: Pandas doesn't allow columns to be created via a new attribute name
                warnings.simplefilter("ignore")
                self.verbose_labels = {x: x for x in self.columns}  # By default, labels are the same as the column names

    # def set_verbose_labels(self, new_verbose_labels: dict) -> None:
    #     """Modify labels"""
    #     self.verbose_labels.update(new_verbose_labels)
    #
    # def clear_verbose_labels(self):
    #     """Clear all labels i.e., set labels to column name"""
    #     self.verbose_labels = {x: x for x in self.columns}  # By default, labels are the same as the column names

    ############################################################################
    # methods which enhance the utility of existing DataFrame methods
    ############################################################################
    def drop(self, labels=None, axis=0, index=None, columns=None, level=None, inplace=False, errors="raise"):
        try:
            return super().drop(labels, axis, index, columns, level, inplace, errors)
        except KeyError as e:
            if str(e)[-18:-1] == "not found in axis":  # make sure it's not some other KeyError
                if labels and axis == 0:
                    labels = search_iterable(self.index, labels)
                elif labels and axis == 1:
                    labels = search_iterable(self.columns, labels)
                elif index:
                    index = search_iterable(self.index, index)
                elif columns:
                    columns = search_iterable(self.columns, columns)
                else:
                    raise KeyError(e)
            return super().drop(labels, axis, index, columns, level, inplace, errors)

    ############################################################################
    # New methods for data wrangling
    ############################################################################
    def keep(self, labels=None, axis=0, index=None, columns=None, level=None, inplace=False, errors="raise"):
        try:
            # test is labels, index, columns are accepted per pandas.DataFrame.drop
            super().drop(labels, axis, index, columns, level, inplace, errors)
            # if it is acceptable, then invert the list
            if labels and axis == 0:
                if type(labels) == str: labels = [labels]
                labels = [x for x in self.index if x not in labels]
            elif labels and axis == 1:
                if type(labels) == str: labels = [labels]
                labels = [x for x in self.columns if x not in labels]
            elif index:
                if type(index) == str: index = [index]
                index = [x for x in self.index if x not in index]
            elif columns:
                if type(columns) == str: columns = [columns]
                columns = [x for x in self.columns if x not in columns]
            else:
                raise KeyError(e)
            return super().drop(labels, axis, index, columns, level, inplace, errors)  # test is labels, index, columns are accepted per pandas.DataFrame.drop
        except KeyError as e:
            if str(e)[-18:-1] == "not found in axis":  # make sure it's not some other KeyError
                if labels and axis == 0:
                    labels = search_iterable(self.index, labels, subtract=True)
                elif labels and axis == 1:
                    labels = search_iterable(self.columns, labels, subtract=True)
                elif index:
                    index = search_iterable(self.index, index, subtract=True)
                elif columns:
                    columns = search_iterable(self.columns, columns, subtract=True)
                else:
                    raise KeyError(e)
            return super().drop(labels, axis, index, columns, level, inplace, errors)

    def gsort(self, varlist):
        """Implements Stata's gsort command
        Note Stata's different interpretation of the minus sign here
        Hence _parse_varlist isn't used

        >>> df.gsort('origfico -origltv')
        """
        df = self  # originally written as a function
        if type(varlist) == str:
            varlist = varlist.split()
        ascending_list =[]
        new_varlist = []
        for x in varlist:
            if x[:1]=='-':
                new_varlist.append(x[1:])
                ascending_list.append(False)
            elif x[:1]=='+':
                new_varlist.append(x[1:])
                ascending_list.append(True)
            else:
                new_varlist.append(x)
                ascending_list.append(True)
        return df.sort_values(new_varlist, ascending=ascending_list)  # .reset_index(drop=True)

    def order(self, varlist, refvar=None, after=True, last=False):
        """
        For whatever reason, it isn't that simple to order columns in pandas.
        This method corrects that.  Note: This currently only supports column
        ordering.

        Parameters
        ----------
        varlist : single label or list-like or magicpandas search string
            The columns to order.
        refvar str, default None
            A reference column e.g., to order after or before refvar
        after : bool, default True
            If True, order after the refvar
        last : bool, default False
            If True, place ordered columns at the end of the dataframe

        Returns
        -------
        MagicDataFrame
            DataFrame with ordered columns.
        """
        df = self
        if type(varlist) == str:
            varlist = varlist.split()
        # varlist = df._parse_varlist(varlist)
        varlist = search_iterable(df, varlist)

        assert type(after) == bool
        assert type(last) == bool
        assert last is False or refvar is None, \
            'last and refvar options cannot be use simultaneously'

        therest = [x for x in list(df.columns) if x not in varlist]

        if refvar is None:
            if len(therest) != 0:
                refvar = therest[0]
            after = False

        if last is True:
            after = True
            refvar = therest[-1]

        if len(therest) != 0:
            colnum = list(therest).index(refvar)
        else:
            colnum = 0

        if after is False:
            head = therest[:colnum]
            tail = therest[colnum:]
        else:
            head = therest[:colnum+1]
            tail = therest[colnum+1:]

        newlist = head
        newlist.extend(varlist)
        newlist.extend(tail)
        df = df[newlist]
        return df

    def weighted_average(self, weight=None, group_by=None):
        df = self.copy()

        if weight:
            if weight not in self.columns:
                raise ValueError(f'"f{weight}" is not a column in DataFrame object.')
            elif not self[weight].dtype == 'float' and not self[weight].dtype == 'int':
                raise TypeError(f'Column "{weight}" must be a float or int data type.')
            elif not isinstance(weight, str):
                raise TypeError(f'"weight" must be a string and refer to a column name.')
        else:
            weight = '__weight__'
            df[weight] = 1

        if group_by:
            if group_by not in self.columns:
                raise ValueError(f'"{group_by}" is not a column in DataFrame object.')
            elif self[group_by].dtype == 'float' or self[group_by].dtype == 'int':
                raise TypeError(f'Column "{group_by}" cannot be a float or int data type.')
            elif not isinstance(group_by, str):
                raise TypeError(f'"group_by" must be a string and refer to a column name.')
        else:
            group_by = '__groupby__'
            df[group_by] = ''

        weight_and_groupby_df = df[[weight, group_by]]
        weight_and_groupby_df = weight_and_groupby_df.groupby(group_by).sum()
        df = pd.merge(df, weight_and_groupby_df, how='left',
                      left_on=group_by, right_index=True,
                      suffixes=('', '_groupby_sum'))
        df[weight] = df[weight] / df[f"{weight}_groupby_sum"]
        del df[f"{weight}_groupby_sum"]
        weight_series = df[weight]
        groupby_series = df[group_by]
        df = df.drop([weight, group_by], axis=1)
        df_times_weights = df.apply(lambda x: x*weight_series)
        df_times_weights[group_by] = groupby_series
        final_grouped = df_times_weights.groupby(group_by)
        # df = df_times_weights.groupby(group_by).sum()
        df = final_grouped.sum()

        # cleanup
        if group_by == '__groupby__':
            df.index.name = ''

        # print(self.verbose_labels)
        # print(self.verbose_labels)
        # print(self.verbose_labels)
        # return MagicDataFrame(df, magic_frame_to_emulate=self)  # needed because groupby doesn't preserve object attributes
        return df

    # def srecode(self, varname, binsizes, right_columns, pyformats):
    #     """Implements Sean's srecode in Python
    #     Note: inplace=True is the default
    #
    #     step_size: default step_size
    #     """
    #     # TODO: convert to inplace=False to conform to pandas
    #     df = self
    #
    #     if varname in right_columns:
    #         right = True
    #     else:
    #         right = False
    #
    #     step_size = binsizes[varname]
    #
    #     fmt = pyformats[varname]  # e.g., {:3.2f}
    #     thisformat = fmt.replace('{:', '{0:') + ' - ' + fmt.replace('{:', '{1:')  # e.g., "{0:3.1f} - {1:3.1f}"
    #     delta = 1/10**int(re.findall(r'\.\d+', fmt)[0].replace('.', ''))  # e.g., 0.01
    #
    #     minvalue = int(df[varname].min()/step_size)*step_size
    #     maxvalue = int(df[varname].max()/step_size+1)*step_size
    #
    #     if right:
    #         bin_labels = [thisformat.format(i+delta, i+step_size) for i in np.arange(minvalue, maxvalue, step_size)]
    #         bin_labels[0] = thisformat.format(minvalue, minvalue+step_size)
    #     else:
    #         bin_labels = [thisformat.format(i, i+step_size-delta) for i in np.arange(minvalue, maxvalue, step_size)]
    #
    #     df['_'+varname] = pd.cut(self[varname], np.arange(minvalue, maxvalue+step_size, step_size), right=right, labels=bin_labels, include_lowest=True)
    #     # self['_' + varname] = self['_'+varname].cat.remove_unused_categories()
    #     # varlbl = labels.get(varname, None)
    #     # if varlbl :
    #     #     labels['_'+varname] = varlbl
    #     return df

    ############################################################################
    # New methods that support finance calculations e.g., IRR
    ############################################################################
    def bc35_price_gen(self, beyield='beyield', coupon='coupon', cpr='cpr',
                       age='age', origfixedterm='origfixedterm',
                       origterm='origterm', dayofmonth=1, delay=47,
                       servicing_fee='servicing_fee', price='price',
                       moodurn='moddurn', wal='wal'):
        cleanprice, adjmoddurn, adjwal = \
            bc35_price(beyield=self[beyield], coupon=self[coupon],
                       cpr=self[cpr], age=self[age],
                       origfixedterm=self[origfixedterm],
                       origterm=self[origterm], dayofmonth=dayofmonth,
                       delay=delay, servicing_fee=self[servicing_fee])
        df = self.copy()
        df[price] = MagicSeries(cleanprice)
        df[moodurn] = MagicSeries(adjmoddurn)
        df[wal] = MagicSeries(adjwal)
        return df

    def bc35_yield_gen(self, price='price', coupon='coupon', cpr='cpr',
                       age='age', origfixedterm='origfixedterm',
                       origterm='origterm', dayofmonth=1, delay=47,
                       servicing_fee='servicing_fee', beyield='beyield'):
        beyield_array = \
            bc35_yield(price=self[price], coupon=self[coupon],
                       cpr=self[cpr], age=self[age],
                       origfixedterm=self[origfixedterm],
                       origterm=self[origterm], dayofmonth=dayofmonth,
                       delay=delay, servicing_fee=self[servicing_fee])
        df = self.copy()
        df[beyield] = MagicSeries(beyield_array)
        return df
    ############################################################################
    # New methods that support I/O and visualization
    ############################################################################
    def browse(self, path: str = None, limit: int = None, run: bool = True,
               percentage_columns: list = None, autofit_columns: list = None,
               client: str = 'excel') -> None:
        if client == "excel":
            easy_xlsxwriter(self, path, limit, run, percentage_columns, autofit_columns)
        elif client == "webbrowser":
            # UNDER CONSTRUCTION
            from magicpandas.chromify import chromify
            chromify(self.to_html(), path, run)
        else:
            raise ValueError('Argument "client" must be either excel or webbrowser')

    def inspect_for_django(self, model_name='DjangoModel'):
        result = ''
        for label, verbose_label in self.verbose_labels.items():
            try:
                if self.dtypes[label] == 'object':  # i.e., a string
                    result += f"""    {label} = models.CharField(max_length=255, verbose_name="{verbose_label}", null=True)\n"""
                elif self.dtypes[label] == 'float64':
                    result += f"""    {label} = models.FloatField(verbose_name="{verbose_label}", null=True)\n"""
                elif self.dtypes[label] == 'datetime64[ns]':
                    result += f"""    {label} = models.DateField(verbose_name="{verbose_label}", null=True)\n"""
                elif self.dtypes[label] == 'bool':
                    result += f"""    {label} = models.BooleanField(verbose_name="{verbose_label}", null=True)\n"""
                elif self.dtypes[label] == 'int64':
                    result += f"""    {label} = models.BigIntegerField(verbose_name="{verbose_label}", null=True)\n"""
                else:
                    raise Exception(f"{label}: There is currently no support for type {self.dtypes[label]}")
            except KeyError as e:
                import warnings
                warnings.warn(f"{label} appears to be in the key file but not in the data file.", stacklevel=2)
        print(f'class {model_name}(models.Model):')
        print(result)

    def to_django(self, DjangoModel, if_exists='fail'):
        """Uses bulk_create to insert data to Django table
        if_exists: see pd.DataFrame.to_sql API

        Ref: https://www.webforefront.com/django/multiplemodelrecords.html
        """
        # TODO: How to I test this?
        if not if_exists in ['fail', 'replace', 'append']:
            raise Exception('if_exists must be fail, replace or append')

        if if_exists=="replace":
            DjangoModel.objects.all().delete()
        elif if_exists=="fail":
            if DjangoModel.objects.all().count()>0:
                raise ValueError('Data already exists in this table')
        else:
            pass

        dct = self.replace({np.nan: None}).to_dict('records') # replace NaN with None since Django doesn't understand NaN

        bulk_list = []
        for x in dct:
            bulk_list.append(DjangoModel(**x))
        DjangoModel.objects.bulk_create(bulk_list)
        print('Successfully saved DataFrame to Django table.')

    def get_dtype_dict(self):
        """
        Returns
        -------
        dict
            Return a dictionary whose keys are column types and values are a
            list containing column labels corresponding to that data type.

        Note: orginally developed for graph(), but no longer used
        """
        dtype_dict = {}
        column_types = self.dtypes.apply(lambda x: str(x)).to_dict()
        for column_label, column_type in column_types.items():
            if dtype_dict.get(column_type, None):
                list_of_columns = dtype_dict.get(column_type, None)
                list_of_columns.append(column_label)
                dtype_dict.update({column_type: list_of_columns})
            else:
                dtype_dict.update({column_type: [column_label]})
        return dtype_dict

    def graph(self, columns=None, mark='circle', facet_columns=1000, run=True):
        """
        Inspired by https://www.youtube.com/watch?v=vTingdk_pVM&t=1260s
        Imposes a clear hierarchy of encodings to make graphing as simple as df.graph()

        Five encodings are supported:
        1) x axis (Q or N)
        2) y axis (Q or N)
        3) facet (Q or N)
        4) size (Q or N)
        5) color (Q or N)
        """
        if not columns:
            columns = list(self.columns)
        if type(columns) == str:
            columns = search_iterable(self.columns, columns)
        df = self[columns].rename(columns=self.verbose_labels)

        encoding_dict = {}
        encoding_list = ['x', 'y', 'facet', 'size', 'color']
        column_list = list(df.columns)
        for i in range(len(column_list)):
            if len(encoding_list) > 0:
                this_encoding = encoding_list.pop(0)
                this_column = column_list.pop(0)
                if this_encoding == "facet":
                    encoding_dict.update({this_encoding: alt.Facet(this_column, columns=facet_columns)})
                elif this_encoding == "x":
                    encoding_dict.update({this_encoding: alt.X(this_column, scale=alt.Scale(zero=False))})
                elif this_encoding == "y":
                    encoding_dict.update({this_encoding: alt.Y(this_column, scale=alt.Scale(zero=False))})
                else:
                    encoding_dict.update({this_encoding: this_column})
            else:
                encoding_dict.update({'tooltip': column_list})
        print(encoding_dict)

        chart = alt.Chart(self)
        chart = getattr(chart, f'mark_{mark}')().encode(**encoding_dict).interactive()
        chromify(chart.to_html(), run=run)

    ############################################################################
    # Shortcut methods e.g., kord = keep & order
    ############################################################################
    def kord(self, varlist):  # , browser_sync=True):
        # keep & order
        df = self
        if type(varlist) == str:
            varlist = varlist.split()
        df = df.keep(varlist, axis=1).order(varlist)  # .reset_index(drop=True)
        return df



import os
import subprocess

def chromify(html_string: str, path=None, run=True):
    if not path:
        import tempfile
        path = os.path.join(tempfile.gettempdir(), 'magicpandas_chromify.html')
    with open(path, 'w') as fn:
        fn.write(html_string)
    if run:
        subprocess.run(f"start chrome {path}", shell=True)



# from pprint import pprint; pprint(locals()); assert False
from magicpandas.finance.bc35 import bc35_price, bc35_yield
import numpy as np


def test_that_bc35_price_is_invertible_with_scalars():
    scalar_price = bc35_price(6, 5, 15, 0, 360, 360)[0]
    scalar_yield = bc35_yield(scalar_price, 5, 15, 0, 360, 360)
    assert abs(scalar_yield - 6) < 0.000000001


def test_that_bc35_price_works_with_numpy_arrays():
    scalar_result_0 = bc35_price(6, 5, 15, 0, 360, 360)
    scalar_result_1 = bc35_price(7, 5, 15, 0, 360, 360)
    beyield = np.array([6, 7])
    coupon = np.array([5, 5])
    cpr = np.array([15, 15])
    age = np.array([0, 0])
    origfixedterm = np.array([360, 360])
    origterm = np.array([360, 360])
    array_result = bc35_price(beyield, coupon, cpr, age, origfixedterm, origterm)
    assert scalar_result_0[0] == array_result[0][0]
    assert scalar_result_0[1] == array_result[1][0]
    assert scalar_result_0[2] == array_result[2][0]
    assert scalar_result_1[0] == array_result[0][1]
    assert scalar_result_1[1] == array_result[1][1]
    assert scalar_result_1[2] == array_result[2][1]


def test_that_bc35_price_is_invertible_with_numpy_arrays():
    beyield = np.array([5, 6])
    coupon = np.array([5, 5])
    cpr = np.array([15, 15])
    age = np.array([0, 0])
    origfixedterm = np.array([360, 360])
    origterm = np.array([360, 360])
    price_array = bc35_price(beyield, coupon, cpr, age, origfixedterm, origterm)
    yield_array = bc35_yield(price_array[0], coupon, cpr, age, origfixedterm, origterm)
    assert abs(yield_array[0] - 5) < 0.000000001
    assert abs(yield_array[1] - 6) < 0.000000001



from time import time
import numpy as np
from numpy.testing import assert_almost_equal

from magicpandas.finance.cashflows import mortgage_cashflows
from magicpandas.finance.irr import pv, irr
from time import time
from magicpandas.finance.irr import pv, irr, mortgage_price, mortgage_yield
from magicpandas.finance.bc35 import bc35_price, bc35_yield
from magicpandas.finance.cashflows import mortgage_cashflows


def test_pv_v1():
    t0 = time()
    pmt = mortgage_cashflows(5, 360, 100000, smm=0.02, pmt_only=True)
    pv2 = pv(5/1200, pmt)
    assert abs(pv2-100000) < 0.0000001
    print('[{:,.2f} ms] '.format(1000*(time()-t0)), end='')


def test_irr_v1():
    t0 = time()
    pmt = mortgage_cashflows(5, 360, 100000, smm=0.02, pmt_only=True)
    irr2 = irr(100000, pmt)
    assert abs(irr2*1200-5) < 0.0000001
    print('[{:,.2f} ms] '.format(1000*(time()-t0)), end='')


def test_irr_numpy_financial():
    t0 = time()
    pmt = mortgage_cashflows(5, 360, 100000, smm=0.02, pmt_only=True)
    irr2 = irr(100000, pmt, use_scipy_optimize_root=False)
    assert abs(irr2*1200-5) < 0.0000001
    print('[{:,.2f} ms] '.format(1000*(time()-t0)), end='')


def test_mortgage_price_v1():
    t0 = time()
    pmt = mortgage_cashflows(5, 360, 100000, smm=0.02, pmt_only=True)
    yld = 200 * ((1 + 0.06 / 12) ** 6 - 1)
    v1 = mortgage_price(yld, 5, 100000, pmt, delay=30)
    cpr = 100*(1-(1-0.02)**12)
    v2, x, y = bc35_price(yld, 5, cpr, 0, 360, 360, delay=30)
    assert abs(v1-v2) < 0.000001
    print('[{:,.2f} ms] '.format(1000*(time()-t0)), end='')


def test_bond_yield_v1():
    t0 = time()
    pmt = mortgage_cashflows(5, 360, 100000, smm=0.02, pmt_only=True)
    cpr = 100*(1-(1-0.02)**12)
    v1 = mortgage_yield(101, 5, 100000, pmt, delay=44)
    v2 = bc35_yield(101, 5, cpr, 0, origfixedterm=360, origterm=360, delay=44)
    assert abs(v1-v2) < 0.000001
    print('[{:,.2f} ms] '.format(1000*(time()-t0)), end='')



"""
bc35 is named after a Bloomberg terminal calculator of the same name.

irr.py: Perform IRR calculations on as arbitrary set of cash flows.
cashflows.py: Create mortgage cash flows (pools not bonds) based on a small set
    of scalars.
bc35.py: Combine irr.py and cashflows.py i.e., perform IRR calculations on
    mortgage cash flows, which are defined based on a small number of scalars.

It turns out that BC35 calculations can be performed using a closed form
    formula.  This calculation is alternative method that can be used to improve
    execution time as well as for unit testing.
"""

import numpy as np
import numpy_financial as npf
from scipy import optimize
import warnings

'''
# BC35
def bc35_price(yld, rate, nper, upb, cpr, servicing_fee=0.0):
    # UNDER CONSTRUCTION
    smm = 1 - (1-cpr/100)**(1/12)
    pmt = pmt_array(rate/1200, nper, -upb, smm=smm, servicing_fee=servicing_fee/1200) # , current_pandi=None, return_frame=False)
    return price
'''


def bc35_price(beyield, coupon, cpr, age, origfixedterm, origterm, dayofmonth=1,
               delay=47, servicing_fee=0, root=None):  # 47 is market standard according to Goldmach Sachs
    '''
    This implements BC35 in core Python (vs columnwise in a pandas.DataFrame)
    '''
    epsilon = servicing_fee/1200

    r = coupon/1200
    y = ((1+beyield/200)**(1/6)-1)
    s = (1-(1-cpr/100)**(1/12))

    m = (origfixedterm - age)
    n = (origterm - age)

    qn = (1 - 1/(1+r)**n)
    l1 = (r*(1-s)-epsilon+s*(1+r))/qn  												# lambda 1
    l2 = (s*(1+r)-epsilon)*(1-1/qn)													# lambda 2
    l3 = ((1-1/qn)*((1+r)*(1-s))**m+(1/qn)*(1-s)**m)							# lambda 3

    pvl1 = (l1/(y+s))*(1-((1-s)/(1+y))**m)							# PV of lamda 1 term
    pvl2 = (l2/(y-(1+r)*(1-s)+1))*(1-((1+r)*(1-s)/(1+y))**m)		# PV of lamda 2 term
    pvb = (l3/(1+y)**m)													# PV of the balloon term
    price = 100*(pvl1+pvl2+pvb)

    # Mod Durn
    dvl1 = (l1/(y+s))*((m/(1+y))*((1-s)/(1+y))**m)-(l1/(y+s)**2)*(1-((1-s)/(1+y))**m)
    dvl2 = (l2/(y-(1+r)*(1-s)+1))*((m/(1+y))*((1+r)*(1-s)/(1+y))**m)-(l2/(y-(1+r)*(1-s)+1)**2)*(1-((1+r)*(1-s)/(1+y))**m)
    dvb = (-m*l3/(1+y)**(m+1))
    adjfactor = (12*(1+beyield/200)**(5/6))
    staticdv01 = -(dvl1 + dvl2 + dvb)/adjfactor  # adjustment factor to reflect derivative based on bond equivalent yield (using the chain rule)
    moddurn = staticdv01/(price/100)

    # WAL
    l1 = (r*(1-s)+s*(1+r))/qn  												# HACK: lambda 1 assuming no servicing fee; price/moddurn fine withou the servicing fee adjustment, but the WAL math breaks
    l2 = (s*(1+r))*(1-1/qn)													# HACK: lambda 2 assuming no servicing fee; price/moddurn fine withou the servicing fee adjustment, but the WAL math breaks
    multl1 = (l1-r/qn)
    wall1 = ((1-((1-s))**(m+1))/(1-((1-s)))-(m+1)*((1-s))**m)/(1-((1-s)))
    multl2 = (l2+(r/qn)/(1+r)**n)
    wall2 = ((1-((1-s)*(1+r))**(m+1))/(1-((1-s)*(1+r)))-(m+1)*((1-s)*(1+r))**m)/(1-((1-s)*(1+r)))

    qnlessm = (1 - 1/(1+r)**(n-m))
    walb = (m*qnlessm/qn*(1-s)**(m))

    wal = (multl1*wall1+multl2*wall2+walb)/12  # wal1+wal2+walb

    # day and delay adjustments
    forwardprice = price*(1+y)**((dayofmonth-1+30-delay)/30)
    cleanprice = forwardprice - coupon*(dayofmonth-1)/360

    waladjustment = (dayofmonth-1+30-delay)/360
    adjwal = wal - waladjustment
    adjmoddurn = moddurn - waladjustment/(1+beyield/200)

    if root is None:
        return (cleanprice, adjmoddurn, adjwal)
    else:
        return cleanprice - root


def bc35_yield(price, coupon, cpr, age, origfixedterm, origterm, dayofmonth=1,
               delay=47, servicing_fee=0):
    # 47 is the market standard according to Goldman Sachs' loan trader
    result = optimize.root(bc35_price, coupon,
                           args=(coupon, cpr, age, origfixedterm,
                                 origterm, dayofmonth, delay,
                                 servicing_fee, price), method='df-sane')

    try:
        # if result.x is a scalar
        # if result.x is np.array or pd.Series, then float returns a TypeError
        return float(result.x)
    except TypeError:
        return result.x



import re
import pandas as pd
from typing import Iterable


def _search_lazily(list_to_search: Iterable, search_string: str) -> list:
    """
    Matches columns starting with a string if it's unique in the set of column names.
    For example, 'firstn' would match 'firstname'.  However, if 'firstname' and
    'firstnumber' are two columns, then a NameError is raised.

    :param list_to_search: pandas.DataFrame
    :param search_string: Stata-style list i.e., string with column names separated by a space
    :return: match list

    Example:
    >>> match_list = _search_lazily(list_to_search, 'origfix origfic')
    >>> assert match_list = ['origfixedterm', 'origfico']
    """
    collist = list(list_to_search) # valid inputs: pandas.DataFrame, pandas.core.indexes.base.Index, ordinary Python list
    if type(search_string) == str:
        search_string = search_string.split()

    matchlist = []
    for var in search_string:
        try:
            collist.index(var)  # produces a ValueError if not found
            matchlist.append(var)
        except ValueError:
            matches = [x for x in collist if x.startswith(var)]
            if len(matches)==1:
                matchlist.append(matches[0])
            elif len(matches)==0:
                raise NameError('variable '+var+' not found')
            else:
                raise NameError(var+' ambiguous abbreviation')

    return matchlist


def _search_by_range(list_to_search: Iterable, search_string: str) -> list:
    """
    Matches columns based on a range, as specified by the syntax "startvar - endvar"

    :param list_to_search: pandas.DataFrame
    :param search_string: Column range, specified by the syntax "startvar - endvar"
    :return: match list

    Example:
    >>> match_list = _search_by_range(list_to_search, 'origfixedterm-origfico')
    >>> assert match_list = ['origfixedterm', 'origltv', 'origfico']
    """
    collist = list(list_to_search) # valid inputs: pandas.DataFrame, pandas.core.indexes.base.Index, ordinary Python list
    if type(search_string) == str:
        search_string = search_string.split()

    search_string = ''.join(search_string)
    search_string = search_string.split('-')

    pos0 = collist.index(search_string[0])
    pos1 = collist.index(search_string[1])
    matchlist = collist[pos0: pos1+1]

    return matchlist


def _search_by_ranges(list_to_search: Iterable, search_string: str) -> list:
    """Handle the case where more than one range is specified"""
    collist = list(list_to_search) # valid inputs: pandas.DataFrame, pandas.core.indexes.base.Index, ordinary Python list
    if type(search_string) == str:
        search_string = search_string.split()

    matchlist = []
    for subsearch_string in search_string:
        thismatchlist = _search_by_range(collist, subsearch_string)
        matchlist.extend(thismatchlist)

    return matchlist


def _search_by_wildcard(list_to_search: Iterable, search_string: str) -> list:
    """
    Matches columns based on * wildcards e.g., "orig*"

    :param list_to_search: pandas.DataFrame
    :param search_string: String containing variables specified by wildcards
    :return: match list
    """
    collist = list(list_to_search) # valid inputs: pandas.DataFrame, pandas.core.indexes.base.Index, ordinary Python list

    if type(search_string) == str:
        search_string = search_string.split()

    relist = []
    for x in search_string:
        x = x.replace('*','.*')
        relist.append(x)
    restr = '^'+'$|^'.join(relist)+'$'
    matchlist = [x for x in collist if re.match(restr,x)]
    return matchlist

def search_iterable(list_to_search: Iterable, search_string: str, subtract: bool=False) -> list:
    """Search any iterable by wildcard, by range, or "lazily"
    """
    collist = list(list_to_search)  # recast iterable as an ordinary list
    if type(search_string) == str:
        search_string = search_string.split()

    typelist = []
    for x in search_string:
        if x.find('*')>-1: typelist.append('w')  # wildcard
        elif x.find('-')>-1: typelist.append('r')  # range
        else: typelist.append('l')  # lazy

    if len(set(typelist))==1 and typelist[0]=='w': matchlist = _search_by_wildcard(list_to_search, search_string)
    elif len(set(typelist))==1 and typelist[0]=='r': matchlist = _search_by_ranges(list_to_search, search_string)
    elif len(set(typelist))==1 and typelist[0]=='l': matchlist = _search_lazily(list_to_search, search_string)
    else: # handle mixed search_string (note: this apparently isn't any slower than a 'pure' search_string)
        matchlist = []
        for x in search_string:
            if x.find('*')>-1:   matchlist.extend(_search_by_wildcard(list_to_search, x))
            elif x.find('-')>-1: matchlist.extend(_search_by_ranges(list_to_search, x))
            else:                matchlist.extend(_search_lazily(list_to_search, x))
    if subtract:
        matchlist = [x for x in collist if x not in matchlist]
    return matchlist



# from pprint import pprint; pprint(locals()); assert False
import pytest
import numpy as np
import pandas as pd
from magicpandas import MagicDataFrame
from magicpandas.finance.bc35 import bc35_price, bc35_yield


@pytest.fixture
def df():
    return MagicDataFrame({
        'coupon': [5, 4.5, 4, 3.5, 3],
        'origfixedterm': [60, 84, 120, 180, 360],
        'origterm': [360, 360, 360, 180, 360],
        'age': [12, 12, 12, 12, 12],
        'cpr': [15, 15, 15, 15, 15],
        'origupb': [100, 200, 300, 400, 500],
        'upb': [100, 200, 300, 400, 500],
        'servicing_fee': [0, 0, 0, 0, 0],
    })


def test_bc35_price_gen(df):
    df['beyield'] = 5
    df = df.bc35_price_gen()
    # assert abs(df._price[0] - 99.901962) < 0.0001
    beyield = np.array(df.beyield)
    coupon = np.array(df.coupon)
    cpr = np.array(df.cpr)
    age = np.array(df.age)
    origfixedterm = np.array(df.origfixedterm)
    origterm = np.array(df.origterm)
    price_array = bc35_price(beyield, coupon, cpr, age, origfixedterm, origterm)[0]
    assert (np.array(df.price) == price_array).all()


def test_bc35_yield_gen(df):
    df['price'] = 99.901962
    df = df.bc35_yield_gen()
    # assert abs(df._price[0] - 99.901962) < 0.0001
    price = np.array(df.price)
    coupon = np.array(df.coupon)
    cpr = np.array(df.cpr)
    age = np.array(df.age)
    origfixedterm = np.array(df.origfixedterm)
    origterm = np.array(df.origterm)
    yield_array = bc35_yield(price, coupon, cpr, age, origfixedterm, origterm)
    assert (np.array(df.beyield) == yield_array).all()



from .fixtures import df1 as df
from magicpandas.utils import search_iterable


def test_wildcard() -> None:
    assert search_iterable(df.columns, 'orig* *upb') == \
           ['origfixedterm',
            'origterm',
            'origltv',
            'origfico',
            'origupb', 'upb']


def test_range() -> None:
    assert search_iterable(df.columns, 'origterm-age') == \
           ['origterm',
            'origltv',
            'origfico',
            'age']


def test_range_two_arguments() -> None:
    assert search_iterable(df.columns, 'rate-origterm age-upb') == \
           ['rate',
            'origfixedterm',
            'origterm',
            'age',
            'origupb',
            'upb']


def test_lazily() -> None:
    assert search_iterable(df.columns, 'origfix origfic') == \
           ['origfixedterm', 'origfico']


def test_mixed() -> None:
    assert search_iterable(df.columns, 'rate ag origterm-origfico *upb') == \
           ['rate', 'age', 'origterm', 'origltv', 'origfico', 'origupb', 'upb']


def test_subtract() -> None:
    assert search_iterable(df.columns, 'rate ag origterm-origfico *upb', subtract=True) == \
           ['origfixedterm']



from .fixtures import df1 as df
from magicpandas.utils import _search_by_wildcard


def test_basic_example():
    assert _search_by_wildcard(df.columns, 'orig*') == \
           ['origfixedterm',
            'origterm',
            'origltv',
            'origfico',
            'origupb']


def test_two_arguments():
    assert _search_by_wildcard(df.columns, 'orig* *upb') == \
           ['origfixedterm',
            'origterm',
            'origltv',
            'origfico',
            'origupb', 'upb']



import types
import functools
import pandas
from .frame import MagicDataFrame, MagicSeries


def magicpandify(func):
    """If an imported function returns DataFrame/Series, return
    a MagicDataFrame/MagicSeries instead"""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        pandas_return_value = func(*args, **kwargs)
        if isinstance(pandas_return_value, pandas.DataFrame):
            return MagicDataFrame(pandas_return_value)
        elif isinstance(pandas_return_value, pandas.Series):
            return MagicSeries(pandas_return_value)
        else:
            return pandas_return_value
    return wrapper


def decorate_all_in_module(module, decorator):
    for name in dir(module):
        obj = getattr(module, name)
        if isinstance(obj, types.FunctionType):
            setattr(module, name, decorator(obj))


decorate_all_in_module(pandas, magicpandify)
# 3 options with base classes: monkey patch or delete or leave unchanged
# pandas.DataFrame = MagicDataFrame
# pandas.Series = MagicSeries
# del pandas.DataFrame
# del pandas.Series
from pandas import *  # This has to come after decorate_all_in_module



import numpy as np
# import pandas as pd  # pandas is only used to collate numpy arrays for summary purposes
# from pandicake.pandaslite import FastFrame
import numpy_financial as npf


def make_smm_array(smm, nper: int) -> np.array:
    """
    Specific the first few smm values to produce a numpy array of length nper.
    This is similar to Bloomberg's custom vectors.

    Parameters
    ----------
    smm : scalar or array_like
        The first
    nper : int
        The number of period of the resulting numpy array

    Returns
    -------
    out : ndarray, float
        A numpy array of length nper
    """
    if type(smm) == list:
        smm = np.array(smm)
    if type(smm) != np.ndarray:  # i.e., float or int
        np_smm = np.repeat(smm, nper)
    elif len(smm) < nper:  # i.e., only the first n periods are specified
        last_value = float(smm[-1:])
        tmp_smm = np.repeat(last_value, nper)
        np_smm = np.append(smm, tmp_smm[len(smm):])
    else:
        np_smm = smm
    return np_smm


def mortgage_cashflows(rate, nper, upb=1000000, smm=1-(1-15/100)**(1/12), sfee=0.0, pandi=None, pmt_only=False):
    """
    Produces an array representing repline cash flows.

    :param rate: The note rate of the mortgage
    :param nper: number of periods e.g., 360 is typical for a new mortgage
    :param upb: current unpaid principal balance
    :param smm: single month mortality rate
    :param current_pandi: optional parameter to specify P&I (for prior curtailments)
    :param return_frame: collates each numpy array and returns a pandas.DataFrame object
    :return: The monthly cash flows of a mortgage repline as a numpy array

    Notes:
    numpy is ~10x faster than the equivalent steps in pandas
    """
    monthly_rate = rate/1200
    monthly_sfee = sfee/1200
    # smm = 1 - (1-cpr/100)**(1/12)  # This is fairly slow if in a vector

    if pandi == None:
        pandi = npf.pmt(rate=monthly_rate, nper=nper, pv=-upb)
    np_smm = make_smm_array(smm, nper)

    np_period = np.arange(nper)
    np_zero_smm_bal = -npf.pv(monthly_rate, nper-np_period, pandi)
    np_period_factor = 1 - np_smm
    np_factor = np_period_factor.cumprod()
    np_factor = np.insert(np_factor, 0, 1)[:-1]
    np_upb = np_factor * np_zero_smm_bal
    np_pandi = npf.pmt(monthly_rate, nper-np_period, -np_upb) - monthly_sfee*np_upb
    np_ipmt = (monthly_rate - monthly_sfee)*np_upb
    np_ppmt = np_pandi - np_ipmt
    np_upmt = (np_upb - np_ppmt)*(1-np_period_factor)
    # np_end_bal = np_upb - np_ppmt - np_upmt  # This should be a test
    np_pmt = np_pandi + np_upmt
    if not pmt_only:
        # np_cpr = 100*(1-(1-np_smm)**12)
        # data = {
        #     # 'smm': np_smm,
        #     'cpr': np_cpr,
        #     'upb': np_upb,
        #     'pmt': np_pmt,
        #     'ipmt': np_ipmt,
        #     'ppmt': np_ppmt,
        #     'upmt': np_upmt,
        # }
        # arr = np.column_stack(tuple(data.values()))
        # return arr
        # return FastFrame(data)
        return np.column_stack((np_upb, np_ipmt, np_ppmt, np_upmt, np_pmt))
    else:
        return np_pmt

################################################################################
# This section is UNDER CONSTRUCTION
# def repline_cashflows(rate, nper, upb, cpr=0.00, sfee=0.0, pandi=None, to_frame=True):
#     return repline_cashflows_monthly(rate/1200, nper, upb,
#             smm=1-(1-cpr/100)**(1/12), monthly_sfee=sfee/1200, pandi=pandi, to_frame=to_frame)

def add_zeros_to_array(array, target_row_count=480):
    """I believe this adds zeros to avoid errors e.g., 30yr loan vs 40yr loan"""
    target_row_count = int(target_row_count)  # error without this
    row_count = np.ma.size(array, axis=0)
    col_count = np.ma.size(array, axis=1)
    additional_rows = target_row_count - row_count
    one_row = np.repeat(0, col_count)
    empty_rows = np.tile(one_row, (additional_rows, 1))
    new_array = np.append(array, empty_rows, axis=0)
    return new_array


def pool_cashflows(rate: np.ndarray, nper: np.ndarray, upb: np.ndarray,
                         smm: np.ndarray, sfee: np.ndarray, pandi: np.ndarray):
    """This produces cash flows for more than 1 loan"""
    # def repline_cashflows(rate, nper, upb=1000000, smm=1-(1-15/100)**(1/12), sfee=0.0, pandi=None, pmt_only=False):
    loan_count = len(rate)
    max_nper = nper.max()
    for j in range(loan_count):
        thispmt = mortgage_cashflows(rate[j], nper[j], upb[j], smm[j], sfee[j], pandi[j])
        thispmt = add_zeros_to_array(thispmt, max_nper)
        if j == 0:
            pmt = thispmt
        else:
            pmt += thispmt
    return pmt



from magicpandas.frame import MagicDataFrame
import warnings
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    import statsmodels.api as sm


def test_keep():
    df = sm.datasets.get_rdataset("Duncan", "carData").data
    df = MagicDataFrame(df)
    df = df.keep('*e', axis=1)
    assert list(df.columns) == ['type', 'income', 'prestige']
    assert len(df.columns) == 3



from .fixtures import df1 as df
from magicpandas.utils import _search_by_ranges


def test_basic_example():
    # Note: Multiple range (e.g., "rate-origterm age-upb" do not work
    assert _search_by_ranges(df.columns, 'origterm-age') == \
           ['origterm',
            'origltv',
            'origfico',
            'age']


def test_two_arguments():
    assert _search_by_ranges(df.columns, 'rate-origterm age-upb') == \
           ['rate',
            'origfixedterm',
            'origterm',
            'age',
            'origupb',
            'upb']



from .fixtures import df1 as df
from magicpandas.utils import _search_lazily


def test_basic_example():
    # unique matches work
    assert _search_lazily(df.columns, 'origfix origfic') == \
           ['origfixedterm', 'origfico']


def test_exact_match():
    assert _search_lazily(df.columns, 'origfixedterm origfico') == \
           ['origfixedterm', 'origfico']


def test_nameerror_ambiguous():
    # with pytest.raises(NameError, message='asdf') as excinfo:
    try:
        _search_lazily(df.columns, 'orig')
    except Exception as e:
        assert type(e) == NameError
        assert str(e) == 'orig ambiguous abbreviation'


def test_nameerror_var_not_found():
    # with pytest.raises(NameError, message='asdf') as excinfo:
    try:
        _search_lazily(df.columns, 'hello')
    except Exception as e:
        assert type(e) == NameError
        assert str(e) == 'variable hello not found'



# Inspiration: https://xlsxwriter.readthedocs.io/working_with_pandas.html
# Note the complexity in handling the header in pandas
import pandas as pd
import subprocess
import os


def easy_xlsxwriter(df: pd.DataFrame, path: str=None, limit: int=None, run: bool=True, percentage_columns: list=None, autofit_columns: list=None) -> None:
    if not path:
        # set default file i.e., tmp/easy_xlsxwriter.xlsx
        import tempfile
        path = os.path.join(tempfile.gettempdir(), 'easy_xlsxwriter.xlsx')

    writer = pd.ExcelWriter(path=path, engine='xlsxwriter')
    if limit:
        df.head(limit).to_excel(writer, sheet_name='Sheet1', startrow=1, header=False)
    else:
        df.to_excel(writer, sheet_name='Sheet1', startrow=1, header=False)
    workbook = writer.book
    worksheet = writer.sheets['Sheet1']

    default_format = workbook.add_format({
        'num_format': '#,##0.00',
        'align': 'center',
        # 'shrink': True,
    })
    for column in df.columns:
        this_column_position = list(df.columns).index(column) + 1
        worksheet.set_column(this_column_position, this_column_position, cell_format=default_format)

    percentage_format = workbook.add_format({
        'num_format': '0%',
        'align': 'center'
    })
    if percentage_columns:
        for column in percentage_columns:
            this_column_position = list(df.columns).index(column) + 1
            worksheet.set_column(this_column_position, this_column_position, cell_format=percentage_format)

    # format header
    header_format = workbook.add_format({
        'bold': True,
        'text_wrap': True,
        'align': 'center',
        'valign': 'top',
        'fg_color': '#D7E4BC',
        'border': 1})
    for col_num, value in enumerate(df.columns.values):
        worksheet.write(0, col_num + 1, value, header_format)
    worksheet.write(0, 0, 'Ticker', header_format)
    worksheet.freeze_panes(1, 1)

    # autofit
    if autofit_columns:
        for column in autofit_columns:
            column_width = df[column].str.len().max()
            this_column_position = list(df.columns).index(column) + 1
            worksheet.set_column(this_column_position, this_column_position, width=column_width)

    # save excel file
    worksheet.set_zoom(75)
    writer.save()
    if run:
        subprocess.run(f'start excel {path}', shell=True)



import pandas_market_calendars as mcal
import numpy as np
from datetime import date
nyse = mcal.get_calendar('NYSE')
holidays = nyse.holidays().holidays


def today_plus(delta):
    """Return number of days from today"""
    date64 = np.datetime64('today')+delta

    if date64.astype(date).isoweekday() == 6:
        date64 += 2
    elif date64.astype(date).isoweekday() == 7:
        date64 += 1

    while date64 in holidays:
        date64 += 1

    return date64.astype(date)



# from magicpandas.utils import search_iterable
from magicpandas.frame import MagicDataFrame
import os
import warnings
from contextlib import suppress
from magicpandas.easy_xlsxwriter import easy_xlsxwriter
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    import statsmodels.api as sm

from contextlib import redirect_stdout
import io

def test_inspect_for_django():
    df = sm.datasets.get_rdataset("Duncan", "carData").data
    df = MagicDataFrame(df)
    f = io.StringIO()
    with redirect_stdout(f):
        df.inspect_for_django()
    correct_output = """class DjangoModel(models.Model):
    type = models.CharField(max_length=255, verbose_name="type", null=True)
    income = models.BigIntegerField(verbose_name="income", null=True)
    education = models.BigIntegerField(verbose_name="education", null=True)
    prestige = models.BigIntegerField(verbose_name="prestige", null=True)

"""
    assert f.getvalue()==correct_output



import sys
import io
from contextlib import _RedirectStream

class capture(_RedirectStream):

    def __init__(self, stream='stdout'):
        self._stream = stream
        self.f = io.StringIO()
        self._new_target = self.f
        # We use a list of old targets to make this CM re-entrant
        self._old_targets = []

    def __enter__(self):
        self._old_targets.append(getattr(sys, self._stream))
        setattr(sys, self._stream, self._new_target)
        return self  # instead of self._new_target in the parent class

    def __repr__(self):
        return self.f.getvalue()



from magicpandas.frame import MagicDataFrame
import warnings
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    import statsmodels.api as sm


def test_drop_columns():
    df = sm.datasets.get_rdataset("Duncan", "carData").data
    df = MagicDataFrame(df)
    df = df.drop('*e', axis=1)  # drop any column that ends with e
    assert df.columns[0] == 'education'
    assert len(df.columns) == 1
